import vrp_c
import supportFunctions.gurobiFunctions as gur_f
from gurobipy import Model, GRB, quicksum
from datetime import datetime
import os
import numpy as np
import pandas as pd
import math

class VRP:
    """
    Класс, в котором объявлены основные параметры, использующиеся в классах CVRP, CVRPTW:
        :type string name_file:   Название файла с данными;
        :type string path_folder: Название папки, в которой буду сохранены бинарные файлы с матрицами расстояний;
        :type int count_towns: Количество городов;
        :type int countTasks:  Количество итераций для решения одной оптимизационной задачи.
    """ 
    def __init__ (self, alg_name, name_file, path_folder, count_towns, shuffle_param):
        self.alg_name    = alg_name
        self.name_file   = name_file 
        self.path_folder = path_folder 
        self.count_towns = count_towns
        self.shuffle_param = shuffle_param
        self.parse_file()


def parse_dist_and_tour(file_name, max_capacity, count_vehicles):
    """
    Функция, возвращающая список из 2 элементов:
        :type float max_count: Длина (в метрах) оптимизированного маршрута;
        :type list mas_dist:  Список подмаршрута оптимизированного маршрута.
    """
    max_count = -1
    mas_dist = []
    with open('CVRP-TW/current_result/res_distance.txt', 'r') as read_file:
        data = read_file.read().split('@')[:-1]
        for i in range(len(data)-1, 0, -2):
            if(float(data[i]) != -1):
                max_count = float(data[i])
                mas_dist = list(map(lambda t: list(map(int, t.split(' '))), data[i-1][:-1].split('#')))
                break
    
    print('Итоговый маршрут: ', mas_dist)
    return [max_count, mas_dist]    

def check(all_pairs, timer, distance, start_lst, finish_lst, service):
    local_timer = timer
    for i, j, _ in all_pairs:
        local_timer += distance[i, j] + service[j]
        if(not(local_timer >= start_lst[j])):
            local_timer = start_lst[j]
        if(not(local_timer >= start_lst[j] and local_timer <= finish_lst[j] and local_timer <= finish_lst[0])):
            return "BAD"
    return local_timer - timer
        
class CVRPTW (VRP):
    """
    Это основной класс для решения задачи CVRPTW
    """
    def __init__ (self, alg_name, name_file, path_folder, count_towns, shuffle_param, capacity: int = 30, time_start: int = 0, time_end: int = 0, count_vehicles: int = 10):
        super().__init__(alg_name, name_file, path_folder, count_towns, shuffle_param)
        self.capacity = capacity
        self.time_start = time_start
        self.time_end   = time_end
        self.count_vehicles = count_vehicles # 20, 50: 10, 100, 200: 20
    
    #TODO: ограничительная функция на все параметры
    
    def parse_file(self):
        """
        Функция, создающая матрицы расстояний, которые записываются в бинарные файлы для задачи CVRPTW. На вход подаются следующие параметры:
            :type string name_file:   Имя файла с данными;
            :type string path_folder: Название папки, в которой буду сохранены бинарные файлы с матрицами расстояний;
            :type int count_towns: Количество городов.
        """
        # для SA и LKH моей реализации
        print("Parse from CVRPTW")
        if(self.alg_name != 'Gurobi'):
            vrp_c.parseOneTwTownPy(self.name_file, self.path_folder, self.count_towns)

    def sa(self) -> [float, list]: #TODO: В SA передавать температуру с клавиатуры
        """
        Функция, вызывающая алгоритм "Имитации отжига" для решения задачи CVRPTW. На вход подается два параметра:
            :type float T:     начальная температура, которая с течением времени убывает;
            :type float t_min: конечная температура, до которой опускается температура T.
        По окончании работы создается файл с данными, состоящий из двух колонок:
            1. В первом столбце записывается длина маршрута, выраженная в метрах, в определенный момент времени;
            2. Во втором столбце записывается время, которое потребовалось, чтобы оптимизировать маршрут до некоторой длины.
        """ 
        print('Введите начальную температуру: ')
        Tstart = float(input())
        print('Введите конечную температуру: ')
        tmin = float(input())
        print('SHUFFLE Param', self.shuffle_param)
        vrp_c.modelMetaHeuristic("cvrptw_sa", self.path_folder, self.count_towns, self.capacity, Tstart, tmin, self.shuffle_param)
        return parse_dist_and_tour(self.name_file, self.capacity, self.count_vehicles)
    def opt(self, name_opt: str = '3opt') -> [float, list]:
        """
        Функция, вызывающая алгоритмы локальной оптимизации для решения задачи CVRPTW. На вход подается один параметр.
            :type string name_opt: название метода оптимизации: 2-opt, 3-opt.
        По окончании работы создается файл с данными, состоящий из двух колонок:
            1. В первом столбце записывается длина маршрута, выраженная в метрах, в определенный момент времени;
            2. Во втором столбце записывается время, которое потребовалось, чтобы оптимизировать маршрут до некоторой длины.
        """
        print('SHUFFLE Param', self.shuffle_param)
        if(name_opt == '2opt'):
            vrp_c.modelMetaHeuristic("cvrptw_2opt", self.path_folder, self.count_towns, self.capacity, 0.0, 0.0, self.shuffle_param)
        elif(name_opt == '3opt'):
            vrp_c.modelMetaHeuristic("cvrptw_3opt", self.path_folder, self.count_towns, self.capacity, 0.0, 0.0, self.shuffle_param)
        return parse_dist_and_tour(self.name_file, self.capacity, self.count_vehicles)

    def lkh(self) -> [float, list]:
        """
        Функция, вызывающая алгоритм "Эвристика Лина-Кёрнигана" для решения задачи CVRPTW.
        По окончании работы создается файл с данными, состоящий из двух колонок:
            1. В первом столбце записывается длина маршрута, выраженная в метрах, в определенный момент времени;
            2. Во втором столбце записывается время, которое потребовалось, чтобы оптимизировать маршрут до некоторой длины.
        """
        print('SHUFFLE Param', self.shuffle_param)
        vrp_c.modelMetaHeuristic("cvrptw_lkh", self.path_folder, self.count_towns, self.capacity, 0.0, 0.0, self.shuffle_param)
        return parse_dist_and_tour(self.name_file, self.capacity, self.count_vehicles)

    def gurobi(self) -> [float, list]:
        """
        Функция, использующая коммерческий проект "Gurobi" для решения задачи CVRPTW.
        
        По окончании работы создается файл с данными, состоящий из двух колонок:
            1. В первом столбце записывается длина маршрута, выраженная в метрах, в определенный момент времени;
            2. Во втором столбце записывается время, которое потребовалось, чтобы оптимизировать маршрут до некоторой длины.
        """ 
        data_file = pd.read_csv(self.name_file, sep="\t", error_bad_lines=True)

        # Запишем все данные в виде массива списков, где каждый список это строка с значениями конкретных столбцов
        arr_data = data_file.values

        # Задаём объем грузовика
        Q_max    = self.capacity # Для 20:500, 50:750, 100:1000

        # Создаём списки координат и грузов для каждого клиента
        xc       = np.zeros(len(arr_data))
        yc       = np.zeros(len(arr_data))
        
        for i in range(len(arr_data)):
            if(arr_data[i][2] <= Q_max):
                xc[i] = arr_data[len(arr_data) - 1 - i][0]
                yc[i] = arr_data[len(arr_data) - 1 - i][1]

        # Создаём список клиентов
        clients  = []
        for i in range(1, len(arr_data)):
            clients.append(i)
        
        # Cоздаем список клиентов вместе с депо. Депо обозначено, как последний элемент в arr_data
        nodes   = [0] + clients

        # Задаём словари временных ограничений на посещение
        
        # Минимальное, максимальное время прибытия + перевод в секунды и время обслуживания
        # Также из списка клиентов удаляются те у кого временное окно меньше времени обслуживания
        start_lst  = {}
        finish_lst = {}
        service    = {}
        q = {}
        del_elem = []
        for i in nodes:
            st = (int(arr_data[len(arr_data) - 1 - i][3].split('-')[0].split(':')[0]) * 60 + int(arr_data[len(arr_data) - 1 - i][3].split('-')[0].split(':')[1])) * 60
            fs = (int(arr_data[len(arr_data) - 1 - i][3].split('-')[1].split(':')[0])  * 60 + int(arr_data[len(arr_data) - 1 - i][3].split('-')[1].split(':')[1])) * 60
            sv = int(arr_data[len(arr_data) - 1 - i][4]) * 60
            if(sv <= fs - st):
                start_lst[i] = st
                finish_lst[i] = fs
                service[i] = sv
                q[i] = arr_data[len(arr_data) - 1 - i][2]
            else:
                del_elem.append(i)

        for i in del_elem:
            clients.remove(i)
            nodes.remove(i)

        # Создаём список всевохможных пар между клиентами, включая депо
        paires  = [(i, j) for i in nodes for j in nodes if i != j]

        # Задаем словарь всевозможных времен необходоимых для перемещения между всеми городами, включая депо
        v = 60 # скорость грузовика в км/ч
        time = {}
        for i, j in paires:
            dist = math.sqrt((xc[i] - xc[j]) ** 2 + (yc[i] - yc[j]) ** 2)
            time[i, j] = dist * 3600 / v

        #Список транспортных средств
        vehicles = [i for i in range(1, self.count_vehicles)]

        #Словарь вместимости транспортных средств
        Q         = {i: Q_max for i in range(1, self.count_vehicles)}

        arco_var  = [(i, j, k) for i in nodes for j in nodes for k in vehicles if i != j]

        arco_time = [(i, k) for i in nodes for k in vehicles]

        # Модель
        model = Model('VRPTW')

        # Переменные для решения
        x = model.addVars(arco_var, vtype=GRB.BINARY, name = 'x')
        t = model.addVars(arco_time, vtype=GRB.CONTINUOUS, name = 't')

        #Целевая функция
        model.setObjective(quicksum(time[i, j] * x[i,j,k] * 60 * 1000 / 3600  for i, j, k in arco_var), GRB.MINIMIZE)

        #Ограничения

        #TODO: Дописать комментарии к ограничениям

        # Прибытие и отъезд
        model.addConstrs(quicksum(x[0, j, k] for j in clients) <= 1 for k in vehicles)
        model.addConstrs(quicksum(x[i, 0, k] for i in clients) <= 1 for k in vehicles)

        # Ограничение по посещению маршрута ровно один раз и ровно одним грузовиком
        model.addConstrs(quicksum(x[i, j, k] for j in nodes for k in vehicles if i != j) == 1 for i in clients)

        #
        model.addConstrs(quicksum(x[i, j, k] for j in nodes if i != j) - quicksum(x[j, i, k] for j in nodes if i != j) == 0 for i in nodes for k in vehicles)

        #Ограничение на вместимость грузовика
        model.addConstrs(quicksum(q[i] * quicksum(x[i, j, k] for j in nodes if i != j) for i in clients) <= Q[k] for k in vehicles)

        #Ограничение по времени
        model.addConstrs((x[i,j,k] == 1) >> (t[i, k] + service[i] + time[i, j] == t[j, k]) for i in clients for j in clients for k in vehicles if i != j)
        model.addConstrs(t[i, k] >= start_lst[i] for i, k in arco_time)
        model.addConstrs(t[i, k] <= finish_lst[i] for i, k in arco_time)

        # Optimizing the model
        model.Params.TimeLimit = 300 # seconds
        model.Params.LogFile= "Gurobi_20_dist.txt"
        model.optimize()
        if model.status == GRB.OPTIMAL:
            print('1.Optimal objective: %g' % model.objVal)
            print('Optimal cost: %g' % model.objVal)
        elif model.status == GRB.INFEASIBLE:
            print('3.Model is infeasible')
        elif model.status == GRB.UNBOUNDED:
            print('4.Model is unbounded')
        else:
            print('5.Optimization ended with status %d' % model.status)
        
        # Список времен в секундах, которое понадобилось для прохождения каждого подмаршрута
        print("model.objVal: ", model.objVal)
        if (model.objVal != math.inf):
            active_arcs = [a for a in arco_var if x[a].x > 0.99]
            res_gur = []
            for c in range(1, self.count_vehicles):
                active_arcs_1 = []
                for i in range(len(active_arcs)):
                    if(active_arcs[i][2] == c):
                        active_arcs_1.append(active_arcs[i])

                res_sort = []
                if(active_arcs_1 != []):
                    res_sort.append(active_arcs_1[0])
                    while(res_sort[len(res_sort)-1][1] != 0):
                        for i in range(len(active_arcs_1)):
                            if(active_arcs_1[i][0] == res_sort[len(res_sort)-1][1]):
                                tu = active_arcs_1[i]
                                break
                        res_sort.append(tu)

                    res_gur.append(check(res_sort, start_lst[0], time, start_lst, finish_lst, service))
            # Общее время прохождения всего маршрута
            s = 0
            count_bad = 0
            r = model.objVal
            for i in res_gur:
                if(i != 'BAD'):
                    s += i
                else:
                    s = "NoSolution"
                    r = "NoSolution"
                    break
            print('')
            print('time Gurobi: ',s)
        else:
            s = "NoSolution"
            r = "NoSolution" 
         
        if(len(arr_data) == 21):
            with open(f"20TownsResult/Result.csv", "a") as f:
                f.write(self.name_file.split('/')[1].split('.csv')[0] + ' ' + str(s) + ' ' + str(model.Runtime) + '\n')
            f.close()
            with open(f"20TownsResult/Result_dist.csv", "a") as f:
                f.write(self.name_file.split('/')[1].split('.csv')[0] + ' ' + str(r) + ' ' + str(model.Runtime) + '\n')
            f.close()
        elif(len(arr_data) == 51):
            with open(f"50TownsResult/Result.csv", "a") as f:
                f.write(self.name_file.split('/')[1].split('.csv')[0] + ' ' + str(s) + ' ' + str(model.Runtime) + '\n')
            f.close()
            with open(f"50TownsResult/Result_dist.csv", "a") as f:
                f.write(self.name_file.split('/')[1].split('.csv')[0] + ' ' + str(r) + ' ' + str(model.Runtime) + '\n')
            f.close()
        elif(len(arr_data)== 101):
            with open(f"100TownsResult/Result.csv", "a") as f:
                f.write(self.name_file.split('/')[1].split('.csv')[0] + ' ' + str(s) + ' ' + str(model.Runtime) + '\n')
            f.close()
            with open(f"100TownsResult/Result_dist.csv", "a") as f:
                f.write(self.name_file.split('/')[1].split('.csv')[0] + ' ' + str(r) + ' ' + str(model.Runtime) + '\n')
            f.close()


from twMethods import *
from collections import deque
import matplotlib.pyplot as plt
import os

def main():
    if(os.path.exists('CVRP-TW/current_result') == False):
        os.mkdir('CVRP-TW/current_result')
    if(os.path.exists('CVRP-TW/results') == False):
        os.mkdir('CVRP-TW/results')
    if(os.path.exists('CVRP-TW/graphs') == False):
        os.mkdir('CVRP-TW/graphs')

    print("Введите количество городов(20, 50, 100): ")
    n = input()
    while(n != '20' and n != '50' and n != '100'):
        print("Введите количество городов(20, 50, 100): ")
        n = input()
    lst = []
    for i in range(1, 4):
        lst.append(f"CVRP-TW/instances/{n}/Example{i}.csv")

    lst_results = []
    lst_legend = []
    print('Если хотиет открыть документацию,то напишите YES в противном случае нажмите Enter')
    doc = input()
    if(doc == 'YES'):
        help(vrp_c) #TODO: в отредактировать текст в документации
    idx = 0
    print('Какой из методов вы хотите использовать для оптимизация маршрута для CVRPTW: SA or LKH or OptAlg or Gurobi?')
    method = input()
    shuffle_param = 0
    i = 0

    open("CVRP-TW/current_result/start_tour.txt", "w")
    meth_alg = ['SA', 'LKH', 'OptAlg', 'Gurobi']
    
    while(1):
        while(i < len(lst)):
            with open(lst[i]) as f:
                count_towns = sum(1 for _ in f) - 1
            with open(lst[i]) as f:
                [last_line] = deque(f, maxlen=1)
                last_line = last_line.split('\t')[3].split('-')
            if(count_towns == 21):
                max_capacity   = 500
                count_vehicles = 11
            elif(count_towns == 51):
                max_capacity   = 750
                count_vehicles = 11
            elif(count_towns == 101):
                max_capacity   = 1000
                count_vehicles = 21
            
            folder_name = f"CVRP-TW/{count_towns-1}_tw"
            if(os.path.exists(folder_name) == False):
                os.mkdir(folder_name)

            if(method == 'SA'):
                start = (int(last_line[0].split(':')[0]) + int(last_line[0].split(':')[1])) * 60
                end   = (int(last_line[1].split(':')[0]) + int(last_line[1].split(':')[1])) * 60
                sa = CVRPTW('SA', lst[i], f"{folder_name}/test{idx}", count_towns, shuffle_param, max_capacity, start, end).sa() #TODO: некоторые параметры брать автоматически из файла
                idx += 1
                with open(f'CVRP-TW/current_result/SA_CVRPTW_result.txt', 'r') as res_file:
                    data = res_file.read()
                    with open(f'CVRP-TW/results/SA_RES_CVRPTW{count_towns-1}.txt', 'a') as write_file:
                        write_file.write(data)
        
                if(f'CVRP-TW/results/SA_RES_CVRPTW{count_towns-1}.txt' in lst_results):
                    lst_results.remove(f'CVRP-TW/results/SA_RES_CVRPTW{count_towns-1}.txt')
                    lst_results.append(f'CVRP-TW/results/SA_RES_CVRPTW{count_towns-1}.txt')
                else:
                    lst_results.append(f'CVRP-TW/results/SA_RES_CVRPTW{count_towns-1}.txt')

            elif(method == 'OptAlg'):
                while(1):
                    print("Input 2opt or 3opt:")
                    name_opt = input()
                    if(name_opt == '2opt' or name_opt == '3opt'):
                        break
                    else:
                        print('Try again!')

                start = (int(last_line[0].split(':')[0]) + int(last_line[0].split(':')[1])) * 60
                end   = (int(last_line[1].split(':')[0]) + int(last_line[1].split(':')[1])) * 60
                opt_alg = CVRPTW('OptAlg', lst[i], f"{folder_name}/test{idx}", count_towns, shuffle_param, max_capacity, start, end).opt(name_opt) #TODO: некоторые параметры брать автоматически из файла
                idx += 1
                if(name_opt == '2opt' or name_opt == '3opt'):
                    with open(f'CVRP-TW/current_result/{name_opt}_CVRPTW_result.txt', 'r') as res_file:
                        data = res_file.read()
                        with open(f'CVRP-TW/results/{name_opt}_RES_CVRPTW{count_towns-1}.txt', 'a') as write_file:
                            write_file.write(data)

                if(f'CVRP-TW/results/{name_opt}_RES_CVRPTW{count_towns-1}.txt' in lst_results):
                    lst_results.remove(f'CVRP-TW/results/{name_opt}_RES_CVRPTW{count_towns-1}.txt')
                    lst_results.append(f'CVRP-TW/results/{name_opt}_RES_CVRPTW{count_towns-1}.txt')
                else:
                    lst_results.append(f'CVRP-TW/results/{name_opt}_RES_CVRPTW{count_towns-1}.txt')

            elif(method == 'LKH'):
                start = (int(last_line[0].split(':')[0]) + int(last_line[0].split(':')[1])) * 60
                end   = (int(last_line[1].split(':')[0]) + int(last_line[1].split(':')[1])) * 60
                lkh = CVRPTW('LKH', lst[i], f"{folder_name}/test{idx}", count_towns, shuffle_param, max_capacity, start, end).lkh() #TODO: некоторые параметры брать автоматически из файла
                idx += 1
                with open(f'CVRP-TW/current_result/LKH_CVRPTW_result.txt', 'r') as res_file:
                    data = res_file.read()
                    with open(f'CVRP-TW/results/LKH_RES_CVRPTW{count_towns-1}.txt', 'a') as write_file:
                        write_file.write(data)
                
                if(f'CVRP-TW/results/LKH_RES_CVRPTW{count_towns-1}.txt' in lst_results):
                    lst_results.remove(f'CVRP-TW/results/LKH_RES_CVRPTW{count_towns-1}.txt')
                    lst_results.append(f'CVRP-TW/results/LKH_RES_CVRPTW{count_towns-1}.txt')
                else:
                    lst_results.append(f'CVRP-TW/results/LKH_RES_CVRPTW{count_towns-1}.txt')

            elif(method == 'Gurobi'):
                gurobi = CVRPTW('Gurobi', lst[i], f"{folder_name}/test{idx}", count_towns, shuffle_param, max_capacity, count_vehicles).gurobi() #TODO: некоторые параметры брать автоматически из файла
                idx += 1
                infile = open('CVRP-TW/results/Gurobi_20.txt',  'r')
                outfile = open('CVRP-TW/results/Gurobi_20_CVRPTW_dist.txt',  'w')

                pst = ['-', 'best', 'continuous,', 'Mon','Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun','second', '']
                copy = False
                k = n
                for line in infile:
                    if 'Expl Unexpl |  Obj  Depth IntInf | Incumbent    BestBd   Gap | It/Node Time' in line:
                        copy = True
                    elif 'Cutting planes:' in line:
                        copy = False
                    elif copy:
                        content = line.split()
                        new_line = ''
                        if len(content) == 10:
                            if((content[5] not in pst) and (prev_content not in pst) and 'e+' not in content[5] and 'e+' not in prev_content):
                                if(float(content[5]) - float(prev_content) > 0.1):
                                    outfile.write('')
                                    
                            if(content[5] not in pst and (prev_content not in pst) and 'e+' not in content[5] and 'e+' not in prev_content):
                                new_line += content[5]
                                prev_content = content[5]
                        elif len(content) == 8:
                            if((content[3] not in pst) and (prev_content not in pst) and 'e+' not in content[3] and 'e+' not in prev_content):
                                if(float(content[3]) - float(prev_content) > 0.1):
                                    outfile.write('')

                            if(content[3] not in pst and (prev_content not in pst) and ('e+' not in content[3]) and ('e+' not in prev_content)):
                                new_line += content[3]
                                prev_content = content[3]
                        if len(content) != 0 and content[-1] != 'processors)' and '.' not in content[-1] and content[-1].isdigit() == False:
                            new_line += '\t' + content[-1] 
                            new_line = new_line[:-1] + '\n'
                            if(new_line[0] != '\t' and new_line[0] != ''):
                                outfile.write(new_line)
                            if(k == 20):
                                if(content[-1][:-1] == '300'):
                                    outfile.write('\n')
                            elif(k == 50):
                                if(content[-1][:-1] == '1000'):
                                    outfile.write('\n')    
                infile.close()
                outfile.close()
                #TODO: добавить сюда сохранение результатов в файл
            if(method not in lst_legend):
                if(method == 'OptAlg'):
                    method = name_opt    
                lst_legend.append(method)

            z = input()
            if(z == ''):
                print('Вы хотите использовать другой алгоритм для оптимизации этой задачи (y/n):')
                z = input()
            while(1):
                if(z == 'n'):
                    break
                elif(z == 'y'):
                    print('Какой из методов вы хотите использовать для оптимизация маршрута для CVRPTW: SA or LKH or OptAlg or Gurobi?')
                    method = input()
                    while(method not in meth_alg):
                        print('Попробуйте ещё раз!')
                        print('Какой из методов вы хотите использовать для оптимизация маршрута для CVRPTW: SA or LKH or OptAlg or Gurobi?')    
                        method = input()
                    shuffle_param = 1
                    break
                else:
                    print('Попробуйте ещё раз!')
                    print('Вы хотите использовать другой алгоритм для оптимизации этой задачи (y/n):')
                    z = input()
            if(z == 'n'):
                print('Вы хотите закончить работу программы (YES/NO):')
                z = input()
                while(1):
                    if(z == 'YES'):
                        break
                    elif(z == 'NO'):
                        print('Какой из методов вы хотите использовать для решения новой задачи оптимизации маршрута для CVRPTW: SA or LKH or OptAlg or Gurobi?')
                        method = input()
                        while(method not in meth_alg):
                            print('Попробуйте ещё раз!')
                            print('Какой из методов вы хотите использовать для решения новой задачи оптимизации маршрута для CVRPTW: SA or LKH or OptAlg or Gurobi?')    
                            method = input()
                        shuffle_param = 0
                        i += 1
                        break
                    else:
                        print('Попробуйте ещё раз!')
                        print('Вы хотите закончить работу программы (YES/NO):')
                        z = input()
            if(z == 'YES'):
                break
        if(z == 'YES'):
            break


    # вычисление средних значений по всем примерам и на основе этого построение графиков каждого из алгоритмов
    for name in lst_results:
        with open(name, 'r') as f:
            file = f.readlines()
        costs = []
        times = []

        cur_cost = []
        cur_time = []

        m = float('-inf')
        for line in file:
            if line == '\n':
                costs.append(cur_cost)
                times.append(cur_time)

                cur_cost = []
                cur_time = []

                continue

            line = line.split('\t')
            cur_cost.append(float(line[0]))
            cur_time.append(float(line[1][:-1]))

            if cur_time[-1] > m:
                m = cur_time[-1]
        step = 1
        m = math.ceil(m)

        time_batched = [[0] * (m // step) for _ in range(len(times))]
        cost_batched = [[float('inf')] * (m // step) for _ in range(len(costs))]

        for num in range(len(costs)):
            cur_time = times[num]
            cur_cost = costs[num]

            for n in range((m // step)):
                if n >= len(cur_time):
                    time_batched[num][n] = time_batched[num][n-1]
                    cost_batched[num][n] = cost_batched[num][n-1]
                    continue

                cost_batched[num][int(n // step)] = min(cost_batched[num][int(n // step)], cur_cost[n])
        cost_batched = np.array(cost_batched)
        plt.plot(cost_batched.mean(0))
        
    plt.legend(lst_legend)
    plt.xlabel('Время оптимизации (секунды)', fontsize=7)
    plt.ylabel('Средняя стоимость (секунды)', fontsize=7)
    plt.title(f'{count_towns-1} points', fontsize=16)

    plt.savefig('CVRP-TW/graphs/All_Algorithms.png')

            
if __name__ == "__main__":
    main()


from setuptools import setup, Extension, find_packages
setup(
      name='CVRP-TW', 
      version='0.0.1',

      author= 'German',
      author_email='1069035F@mail.ru',

      url='https://github.com/Wektor607/CVRPTW-models/tree/main/OwnLibrary',
      
      packages=['CVRP-TW'],
      ext_modules=[Extension(
      'vrp_c', 
      sources=[   
            'CVRP-TW/vrp-main.c', 
            'CVRP-TW/algorithms/SA.c', 
            'CVRP-TW/algorithms/twoOpt.c',
            'CVRP-TW/algorithms/threeOpt.c',
            'CVRP-TW/algorithms/LKH.c',
            'CVRP-TW/mainFunctions/distanceMatrix.c',
            'CVRP-TW/mainFunctions/dataProcessing.c',
            'CVRP-TW/mainFunctions/townActions.c',
            'CVRP-TW/mainFunctions/createObjects.c',
            'CVRP-TW/mainFunctions/logistic.c',
      ])]
)


import numpy as np
import os
def generate_data(n_samples=1, 
                  n_customer=20,
                  service_window=1000,
                  service_duration=10,
                  time_factor=100.0,
                  tw_expansion=3.0):
    
    rnds = np.random

    # sample locations
    dloc = rnds.uniform(size=(n_samples, 1, 2))  # depot location
    nloc = rnds.uniform(size=(n_samples, n_customer, 2))  # node locations
    
    # generate capacities 
    min_capasity = 1
    max_capasity = 42
    capacity = np.minimum(np.maximum(np.abs(np.random.normal(15, 10, size=[n_samples, n_customer])), min_capasity), max_capasity)
    
    # TW start needs to be feasibly reachable directly from depot
    min_t = np.ceil(np.linalg.norm(
        dloc[:, :1] * time_factor - nloc * time_factor, axis=-1)) + 1
    # TW end needs to be early enough to perform service and return to depot until end of service window
    max_t = np.ceil(np.linalg.norm(
        dloc[:, :1] * time_factor - nloc * time_factor, axis=-1) + service_duration) + 1
    
    # horizon allows for the feasibility of reaching nodes / returning from nodes within the global tw (service window)
    horizon = list(zip(min_t, service_window - max_t))
    epsilon = np.maximum(np.abs(rnds.standard_normal(
        [n_samples, n_customer])), 1 / time_factor)
    
    # sample earliest start times a
    a = [rnds.randint(*h) for h in horizon]
    # calculate latest start times b, which is
    # a + service_time_expansion x normal random noise, all limited by the horizon
    # and combine it with a to create the time windows
    tw = [np.transpose(np.vstack((rt,
                                  np.minimum(
                                      rt + tw_expansion * time_factor * sd, h[-1]).astype(int)  # b
                                  ))).tolist()
          for rt, sd, h in zip(a, epsilon, horizon)]
    
    return dloc, nloc, capacity, tw, service_duration

def translate_minute(minute): 
    hour = minute // 60 
    minute %= 60 
    return "%02d:%02d" % (hour, minute)

if __name__ == "__main__":

    # Размерность задачи
    print("Введиите количество городов: ")
    n_customer = int(input())
    if(os.path.exists('instances') == False):
        os.mkdir('instances')
    if(os.path.exists(f'instances/{n_customer}') == False):
        os.mkdir(f'instances/{n_customer}')
    # Время открытия и закрытия депо в часах
    print("Введите начало временного окна депо от 0 до 24 часов в формате XX:XX : ")
    depot_start_time = input()
    print("Введите конец временного окна депо от 0 до 24 часов в формате XX:XX : ")
    depot_end_time = input()

    for i in range(1, 51):
        depot, points, capacity, tw, deliver_time = generate_data(n_customer=n_customer)
        with open(f"instances/{n_customer}/Example{i}.csv", "w") as f:
            f.write('x_coord' + '\t' + 'y_coord' + '\t' + 'Weight' + '\t' + 'DeliveryTimeRange' + '\t' + 'UnloadingTime' + '\n')
            
            for j in range(len(points[0])):
                f.write(str(points[0][j][0] * 100) + '\t' + str(points[0][j][1] * 100) + '\t' + str(capacity[0][j]) 
                        + '\t' + str(translate_minute(tw[0][j][0])) + '-' + str(translate_minute(tw[0][j][1])) + '\t' + str(deliver_time) + '\n')
            
            f.write(str(depot[0][0][0] * 100) + '\t' + str(depot[0][0][1] * 100) + '\t' + str(0) 
                        + '\t' + depot_start_time + '-' + depot_end_time + '\t' + str(0))